#!/usr/bin/python3
"""
This script benchmarks all detectors with both the MIT database and the GU database.
For the GU database it runs it for both the chest strap and Einthoven II with loose cables.
You need to download both the MIT arrhythmia database from: https://alpha.physionet.org/content/mitdb/1.0.0/
and install the GUDB API https://github.com/berndporr/ECG-GUDB
The MIT database needs to be placed below this directory: "../mit-bih-arrhythmia-database-1.0.0/".
"""
import numpy as np
import pathlib
from tester_MITDB import MITDB_test
from tester_GUDB import GUDB_test
from ecgdetectors import Detectors
from multiprocessing import Process

# benchmark the detectors with the MIT DB
do_test_MIT = True

# benchmark the detectors with the GU DB
do_test_GU  = True

def run_GUDB_tests(leads):
    # GUDB database testing
    gu_test = GUDB_test()
    gu_detectors = Detectors(250)
    gu_test.classifer_test_all(tolerance=0, config=leads)

def run_MIT_tests():
    # MIT-BIH database testing
    mit_test = MITDB_test()
    mit_detectors = Detectors(360)
    
    # test single detector
    matched_filter_mit = mit_test.single_classifier_test(mit_detectors.matched_filter_detector, tolerance=0)
    np.savetxt('matched_filter_mit.csv', matched_filter_mit, fmt='%i', delimiter=',')

    # test all detectors on MITDB, will save results to csv, will take some time
    mit_test.classifer_test_all()


if do_test_MIT:
    pmit = Process(target=run_MIT_tests)
    pmit.start()

if do_test_GU:
    pgustrap = Process(target=run_GUDB_tests, args=('chest_strap',))
    pgustrap.start()
    pgucables = Process(target=run_GUDB_tests, args=('loose_cables',))
    pgucables.start()

if do_test_MIT:
    pmit.join()

if do_test_GU:
    pgustrap.join()
    pgucables.join()



import numpy as np
from ecgdetectors import Detectors
from datetime import datetime

det_names = ['two_average', 'matched_filter', 'swt', 'engzee', 'christov', 'hamilton', 'pan_tompkins']
det_fullnames = ['Elgendi', 'Matched', 'Kalidas', 'Engzee', 'Christov', 'Hamilton', 'Pan']

def sort_MIT_annotations(ann):
    beat_labels = ['N', 'L', 'R', 'B', 'A', 'a', 'J', 'S', 'V', 'r', 'F', 'e', 'j', 'n', 'E', '/', 'f', 'Q', '?']
    in_beat_labels = np.in1d(ann.symbol, beat_labels)

    sorted_anno = ann.sample[in_beat_labels]
    sorted_anno = np.unique(sorted_anno)

    return sorted_anno


"""
From the detected R peaks the function works itself backwards to
calculate the median delay the detector introduces. This is used
for benchmarking to compensate for different delays the detctors
introduce.
"""
def calcMedianDelay(detected_peaks, unfiltered_ecg, search_samples):

    r_peaks = []
    window = int(search_samples)

    for i in detected_peaks:
        i = int(i)
        if i<window:
            section = unfiltered_ecg[:i]
            r_peaks.append(np.argmax(section))
        else:
            section = unfiltered_ecg[i-window:i]
            r_peaks.append(window - np.argmax(section))

    m = int(np.median(r_peaks))
    #print("Delay = ",m)
    return m


def evaluate_detector(test, annotation, delay, tol=0):

    test = np.unique(test)
    reference = np.unique(annotation)
    
    TP = 0

    for anno_value in test:
        test_range = np.arange(anno_value-tol-delay, anno_value+1+tol-delay)
        in1d = np.in1d(test_range, reference)
        if np.any(in1d):
            TP = TP + 1
    
    FP = len(test)-TP
    FN = len(reference)-TP 

    return TP, FP, FN


def det_from_name(detector_name, fs):

        detectors = Detectors(fs)
        detectors.engzee_fake_delay = 10

        if detector_name=='two_average':
                return detectors.two_average_detector
        elif detector_name=='matched_filter':
                return detectors.matched_filter_detector
        elif detector_name=='swt':
                return detectors.swt_detector
        elif detector_name=='engzee':
                return detectors.engzee_detector
        elif detector_name=='christov':
                return detectors.christov_detector
        elif detector_name=='hamilton':
                return detectors.hamilton_detector
        elif detector_name=='pan_tompkins':
                return detectors.pan_tompkins_detector
        else:
                raise RuntimeError('invalid detector name!')


def get_time():

        time = str(datetime.now().time())
        time = time[:5]
        time = time.replace(':', '.')
        
        return time



import numpy as np
import pandas as pd
import _tester_utils
import wfdb
import pathlib
import os
from ecgdetectors import Detectors


class MITDB_test:
    """
    Benchmarks detectors with against the MITDB.
    You need to download both the MIT arrhythmia database from: https://alpha.physionet.org/content/mitdb/1.0.0/
    and needs to be placed below this directory: "../mit-bih-arrhythmia-database-1.0.0/"
    """
    def __init__(self):
        current_dir = pathlib.Path(__file__).resolve()
        self.mitdb_dir = str(pathlib.Path(current_dir).parents[1]/'mit-bih-arrhythmia-database-1.0.0')

    def single_classifier_test(self, detector, tolerance=0):
        max_delay_in_samples = 350 / 5
        dat_files = []
        for file in os.listdir(self.mitdb_dir):
            if file.endswith(".dat"):
                dat_files.append(file)
        
        mit_records = [w.replace(".dat", "") for w in dat_files]
        
        results = np.zeros((len(mit_records), 5), dtype=int)

        i = 0
        for record in mit_records:
            progress = int(i/float(len(mit_records))*100.0)
            print("MITDB progress: %i%%" % progress)

            sig, fields = wfdb.rdsamp(self.mitdb_dir+'/'+record)
            unfiltered_ecg = sig[:, 0]  

            ann = wfdb.rdann(str(self.mitdb_dir+'/'+record), 'atr')    
            anno = _tester_utils.sort_MIT_annotations(ann)    

            r_peaks = detector(unfiltered_ecg)

            delay = _tester_utils.calcMedianDelay(r_peaks, unfiltered_ecg, max_delay_in_samples)

            if delay > 1:

                TP, FP, FN = _tester_utils.evaluate_detector(r_peaks, anno, delay, tol=tolerance)
                TN = len(unfiltered_ecg)-(TP+FP+FN)
                
                results[i, 0] = int(record)    
                results[i, 1] = TP
                results[i, 2] = FP
                results[i, 3] = FN
                results[i, 4] = TN

            i = i+1
        
        return results


    def classifer_test_all(self, tolerance=0):

        output_names = ['TP', 'FP', 'FN', 'TN']

        total_records = 0
        for file in os.listdir(self.mitdb_dir):
            if file.endswith(".dat"):
                total_records = total_records + 1

        total_results = np.zeros((total_records, 4*len(_tester_utils.det_names)), dtype=int)

        counter = 0
        for det_name in _tester_utils.det_names:

            print('\n'+det_name)

            result = self.single_classifier_test(_tester_utils.det_from_name(det_name, 360), tolerance=tolerance)
            index_labels = result[:, 0]
            result = result[:, 1:]

            total_results[:, counter:counter+4] = result

            counter = counter+4  

        col_labels = []

        for det_name in _tester_utils.det_names:
                for output_name in output_names:
                    label = det_name+" "+output_name
                    col_labels.append(label)

        total_results_pd = pd.DataFrame(total_results, index_labels, col_labels, dtype=int)            
        total_results_pd.to_csv('results_MITDB'+'.csv', sep=',')

        return total_results_pd



import numpy as np
import pandas as pd
import _tester_utils

from ecgdetectors import Detectors
from ecg_gudb_database import GUDb

class GUDB_test:
    """
    This class benchmarks detectors against the GU database.
    You need to install the API for the GUDB: https://github.com/berndporr/ECG-GUDB
    """
    
    def single_classifier_test(self, detector, tolerance=0, config="chest_strap"):

        max_delay_in_samples = 250 / 3

        total_subjects = GUDb.total_subjects

        results = np.zeros((total_subjects, (4*len(GUDb.experiments))+1), dtype=int)

        for subject_number in range(0, total_subjects):
            progress = int(subject_number/float(total_subjects)*100.0)
            print("GUDB "+config+" progress: %i%%" % progress)

            results[subject_number, 0] = subject_number
            exp_counter = 1
            for experiment in GUDb.experiments:
                
                ecg_class = GUDb(subject_number, experiment)

                anno_exists = False
                if config=="chest_strap" and ecg_class.anno_cs_exists:
                    unfiltered_ecg = ecg_class.cs_V2_V1                   
                    anno = ecg_class.anno_cs
                    anno_exists = True
                elif config=="loose_cables" and ecg_class.anno_cables_exists:
                    unfiltered_ecg = ecg_class.einthoven_II 
                    anno = ecg_class.anno_cables
                    anno_exists = True
                elif config!="chest_strap" and config!="loose_cables":
                    raise RuntimeError("Config argument must be chest_strap or loose_cables!")
                    return results

                if anno_exists:                  

                    r_peaks = detector(unfiltered_ecg)

                    delay = _tester_utils.calcMedianDelay(r_peaks, unfiltered_ecg, max_delay_in_samples)
                    print("delay = ",delay)

                    # there must be a delay in all cases so anything below is a bad sign
                    if delay > 1:

                        TP, FP, FN = _tester_utils.evaluate_detector(r_peaks, anno, delay, tol=tolerance)
                        TN = len(unfiltered_ecg)-(TP+FP+FN)

                        results[subject_number, exp_counter] = TP
                        results[subject_number, exp_counter+1] = FP
                        results[subject_number, exp_counter+2] = FN
                        results[subject_number, exp_counter+3] = TN

                exp_counter = exp_counter+4

        return results


    def classifer_test_all(self, tolerance=0, config="chest_strap"):

        output_names = ['TP', 'FP', 'FN', 'TN']

        total_results = np.zeros((GUDb.total_subjects, 4*len(GUDb.experiments)*len(_tester_utils.det_names)), dtype=int)

        counter = 0
        for det_name in _tester_utils.det_names:

            print('\n'+config+" "+det_name+":")

            result = self.single_classifier_test(_tester_utils.det_from_name(det_name, 250), tolerance=tolerance, config=config)
            result = result[:, 1:]

            total_results[:, counter:counter+(4*len(GUDb.experiments))] = result

            counter = counter+(4*len(GUDb.experiments))        

        index_labels = np.arange(GUDb.total_subjects)
        col_labels = []

        for det_name in _tester_utils.det_names:
            for experiment_name in GUDb.experiments:
                for output_name in output_names:
                    label = det_name+" "+experiment_name+" "+output_name
                    col_labels.append(label)

        total_results_pd = pd.DataFrame(total_results, index_labels, col_labels, dtype=int)            
        total_results_pd.to_csv('results_GUDB_'+config+'.csv', sep=',')

        return total_results_pd



#!/usr/bin/python3
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy.stats as stats
import _tester_utils

def str_join(a, b, c):

    a = np.array(a)
    b = np.array(b)
    c = np.array(c)

    str_array = np.core.defchararray.add(np.core.defchararray.add(a, b), c)

    return str_array


def get_sensivities(results_df, detector_name, experiment=None):

    if experiment!=None:
        tp_col_names = str_join(detector_name+' ', [experiment], ' TP')
        fp_col_names = str_join(detector_name+' ', [experiment], ' FP')
        fn_col_names = str_join(detector_name+' ', [experiment], ' FN')
        tn_col_names = str_join(detector_name+' ', [experiment], ' TN')
        total_tp = (results_df.loc[:, tp_col_names].values)[:,0]
        total_fn = (results_df.loc[:, fn_col_names].values)[:,0]
    else:
        tp_col_names = detector_name+' '+'TP'
        fp_col_names = detector_name+' '+'FP'
        fn_col_names = detector_name+' '+'FN'
        tn_col_names = detector_name+' '+'TN'      
        total_tp = results_df.loc[:, tp_col_names].values
        total_fn = results_df.loc[:, fn_col_names].values

    se = []

    for tp,fn in zip(total_tp,total_fn):
            if (tp + fn) > 0:
                s = tp/(tp+fn)*100.0
                se.append(s)

    return np.array(se)


def get_result(results_df, det_names, experiment=None):
    
    m = []
    s = []
    for det in det_names:
        m.append(np.mean(get_sensivities(results_df, det, experiment)))
        s.append(np.std(get_sensivities(results_df, det, experiment)))

    return np.array(m),np.array(s)



def compare_det_test(results_df, detector_name1, detector_name2, experiment=None):
    se1 = get_sensivities(results_df, detector_name1, experiment)
    if len(se1) < 2:
        return 0
    se2 = get_sensivities(results_df, detector_name2, experiment)
    if len(se2) < 2:
        return 0
    l = min(len(se1),len(se2))
    if (l < 20):
        return None
    #print("1:",se1[:l])
    #print("2:",se2[:l])
    try:
        t,p = stats.wilcoxon(se1[:l],se2[:l])
        return p
    except:
        return None

def compare_cables(results_df1,results_df2,det1,experiment):
    se1 = get_sensivities(results_df1, det1, experiment)
    if len(se1) < 2:
        return 0
    se2 = get_sensivities(results_df2, det1, experiment)
    if len(se2) < 2:
        return 0
    l = min(len(se1),len(se2))
    if (l < 20):
        return None
    try:
        t,p = stats.wilcoxon(se1[:l],se2[:l])
        return p
    except:
        return None


def compare_experiments(results_df1,det1):
    se1 = get_sensivities(results_df1, det1, 'sitting')
    if len(se1) < 2:
        return 0
    se2 = get_sensivities(results_df1, det1, 'jogging')
    if len(se2) < 2:
        return 0
    l = min(len(se1),len(se2))
    if (l < 20):
        return None
    try:
        t,p = stats.wilcoxon(se1[:l],se2[:l])
        return p
    except:
        return None


def single_plot(data, std, y_label, title = None):
    fig, ax = plt.subplots()
    _tester_utils.det_fullnames = ['Elgendi et al', 'Matched Filter', 'Kalidas and Tamil', 'Engzee Mod', 'Christov', 'Hamilton', 'Pan and Tompkins']
    x_pos = np.arange(len(_tester_utils.det_fullnames))

    fig.set_size_inches(10, 7)
    rects1 = ax.bar(x_pos, data, yerr=std, width = 0.65, align='center', alpha=0.5, ecolor='black', capsize=10)
    ax.set_ylim([0,100])
    ax.set_ylabel(y_label)
    ax.set_xlabel('Detector')
    ax.set_xticks(x_pos)
    ax.set_xticklabels(_tester_utils.det_fullnames)

    if title!=None:
        ax.set_title(title)

    plt.tight_layout()

    return rects1


def double_plot(data1, std1, data2, std2, y_label, legend1, legend2, title=None):
    fig, ax = plt.subplots()
    _tester_utils.det_fullnames = ['Elgendi et al', 'Matched Filter', 'Kalidas and Tamil', 'Engzee Mod', 'Christov', 'Hamilton', 'Pan and Tompkins']
    x_pos = np.arange(len(_tester_utils.det_fullnames))

    fig.set_size_inches(10, 7)
    width = 0.4
    rects1 = ax.bar(x_pos, data1, width, yerr=std1, alpha=0.5, ecolor='black', capsize=10)
    rects2 = ax.bar(x_pos+width, data2, width, yerr=std2, alpha=0.5, ecolor='black', capsize=10)
    ax.set_ylim([0,100])
    ax.set_ylabel(y_label)
    ax.set_xlabel('Detector')
    ax.set_xticks(x_pos + width / 2)
    ax.set_xticklabels(_tester_utils.det_fullnames)
    ax.legend((rects1[0], rects2[0]), (legend1, legend2))

    if title!=None:
        ax.set_title(title)

    plt.tight_layout()

    return rects1, rects2


def print_stat(p):
    if p == None:
        print('--- & ',end='')
        return
    s = ""
    if p < 0.05:
        s = "*"
    print('{:03.2f}{} & '.format(p,s),end='')

def print_detector_comparision_table(title,datafile,experiment=None):
    print(title)
    print("      & ",end='')
    for detlong1 in _tester_utils.det_fullnames:
        print(detlong1," & ",end='')
    print("\\\\")
    for det1,detlong1 in zip(_tester_utils.det_names,_tester_utils.det_fullnames):
        print(detlong1," & ",end='')
        for det2,detlong2 in zip(_tester_utils.det_names,_tester_utils.det_fullnames):
            p = compare_det_test(datafile, det1, det2, experiment)
            print_stat(p)
        print("\\\\")
    print()

# GUDB
gudb_cs_results = pd.read_csv('results_GUDB_chest_strap.csv', dtype=int, index_col=0)
gudb_cable_results = pd.read_csv('results_GUDB_loose_cables.csv', dtype=int, index_col=0)

# MITDB
mitdb_results = pd.read_csv('results_MITDB.csv', dtype=int, index_col=0)

experiment_names = ['sitting','maths','walking','hand_bike','jogging']
output_names = ['TP', 'FP', 'FN', 'TN']

print_detector_comparision_table("MIT database",mitdb_results)

print_detector_comparision_table("EINTHOVEN SITTING",gudb_cable_results,'sitting')
print_detector_comparision_table("CHEST STRAP SITTING",gudb_cs_results,'sitting')

print_detector_comparision_table("EINTHOVEN JOGGING",gudb_cable_results,'jogging')
print_detector_comparision_table("CHEST STRAP JOGGING",gudb_cs_results,'jogging')

print("\n\n\n")

print("CHEST STRAP VS EINTH")
for exp1 in ['sitting','jogging']:
    print(exp1,":")
    for det1 in _tester_utils.det_names:
        p = compare_cables(gudb_cable_results,gudb_cs_results,det1,exp1)
        print_stat(p)
    print("\\\\")

print("\n\n\n")

print("LOOSE CABLES: SITTING VS JOGGING")
for det1 in _tester_utils.det_names:
    p = compare_experiments(gudb_cable_results,det1)
    print_stat(p)
print("\\\\")

print("\n\n\n")

print("CHEST STRAP: SITTING VS JOGGING")
for det1 in _tester_utils.det_names:
    p = compare_experiments(gudb_cs_results,det1)
    print_stat(p)
print("\\\\")

print("\n\n\n")

# calculating results
mitdb_avg,mitdb_std = get_result(mitdb_results, _tester_utils.det_names)
print("mitdb:\n",_tester_utils.det_fullnames,"\n",mitdb_avg)
gudb_cs_sitting_avg,gudb_cs_sitting_std = get_result(gudb_cs_results,
                                                     _tester_utils.det_names, 'sitting')
print("chest strap sitting:",gudb_cs_sitting_avg)
gudb_cable_sitting_avg,gudb_cable_sitting_std = get_result(gudb_cable_results,
                                                           _tester_utils.det_names, 'sitting')
print("lose cables sitting:",gudb_cable_sitting_avg)
gudb_cs_jogging_avg,gudb_cs_jogging_std = get_result(gudb_cs_results,
                                                     _tester_utils.det_names, 'jogging')
gudb_cable_jogging_avg,gudb_cable_jogging_std = get_result(gudb_cable_results,
                                                           _tester_utils.det_names, 'jogging')

# plotting
single_plot(mitdb_avg, mitdb_std, 'Sensitivity (%)', 'MITDB')

double_plot(gudb_cs_sitting_avg, gudb_cs_sitting_std,
            gudb_cable_sitting_avg, gudb_cable_sitting_std,
            'Sensitivity (%)', 'Chest Strap', 'Loose Cables', 'GUDB: cable, sitting')

double_plot(gudb_cs_jogging_avg, gudb_cs_jogging_std,
            gudb_cable_jogging_avg, gudb_cable_jogging_std,
            'Sensitivity (%)', 'Chest Strap', 'Loose Cables', 'GUDB: cable, jogging')



plt.show()
